import java.nio.FloatBuffer;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.GL13.*;
import static org.lwjgl.opengl.GL15.*;
import static org.lwjgl.opengl.GL20.*;
import static org.lwjgl.opengl.GL21.*;
import static org.lwjgl.opengl.GL30.*;

public class Print {

	// this quad is used by the string printer
	// 16 columns of 6 rows
    public static float quadVerts[] = {
        // vert                     // UV
          0.0f,   0.0f,  0.0f,		0f,			0f,
         32.0f,   0.0f,  0.0f,      0.0625f,	0f,
         32.0f,  42.0f,  0.0f,      0.0625f,	0.1640625f,
         32.0f,  42.0f,  0.0f,      0.0625f,	0.1640625f,
          0.0f,  42.0f,  0.0f,      0f,			0.1640625f,
          0.0f,   0.0f,  0.0f,      0f,			0f
	};
	
	private int texture;
	private int program;
	
	// shader attribs and uniforms
	private int attr_pos=1;
	private int attr_uv=3;
	private int u_matrix,u_texture,u_offset,u_uvoffset;
	
	private FloatBuffer vertsBuffer;
	private int vbo,vao;
	
	private float charWidth; // width of a single char on the screen
	
	private String fragShaderText =
//(for GLES)			"precision mediump float;\n"+
		"uniform sampler2D u_texture;\n"+
		"uniform vec2 u_uvoffset; \n"+
		"varying vec2 v_frag_uv; \n"+
		"void main() {\n"+
		"	vec2 uv =  v_frag_uv;\n"+
		"	uv.x = uv.x + u_uvoffset.x;\n"+
		"	uv.y = uv.y + u_uvoffset.y;\n"+
		"   gl_FragColor = texture2D(u_texture, uv);\n"+
		"}\n";

	private String vertShaderText =
		"uniform mat4 u_matrix;\n"+
		"uniform vec2 u_offset;\n"+
		"attribute vec4 a_pos;\n"+
		"attribute vec2 a_uv;\n"+
		"varying vec2 v_frag_uv;\n"+
		"\n"+
		"void main() {\n"+
		"   vec4 p=a_pos;\n"+
		"   p.x=p.x+u_offset.x;\n"+
		"   p.y=p.y+u_offset.y;\n"+
		"   gl_Position = u_matrix * p ;\n"+
		"	v_frag_uv = a_uv;\n"+
		"}\n";

	
	Print(String texName, float cWidth, float charHeight) {

		charWidth = cWidth;
		quadVerts[ 5] = quadVerts[10] = quadVerts[15] = charWidth;
		quadVerts[11] = quadVerts[16] = quadVerts[21] = charHeight;

		texture = Util.loadTexture(texName,true);
		
		program = Util.createShaderProgram("font",vertShaderText,fragShaderText);
        glUseProgram(program);
           

        ByteBuffer val = ByteBuffer.allocateDirect(4);
        
        glBindAttribLocation(program, attr_pos,     "a_pos");
        glBindAttribLocation(program, attr_uv,      "a_uv");

        glLinkProgram(program);  /* needed to put attribs into effect */
        glGetProgrami(program, GL_LINK_STATUS, val);
        
        if (val.asIntBuffer().get(0)==0) {
            System.out.println("(font) shader log:\n"+glGetProgramInfoLog(program));
            System.exit(-1);
        }

        u_matrix = glGetUniformLocation(program, 	"u_matrix");
        u_texture = glGetUniformLocation(program, 	"u_texture");		
        u_offset = glGetUniformLocation(program, 	"u_offset");		
        u_uvoffset = glGetUniformLocation(program,	"u_uvoffset");		
        
        
        vertsBuffer = ByteBuffer.allocateDirect(quadVerts.length * 4).order(ByteOrder.nativeOrder())
                                    .asFloatBuffer().put(quadVerts);
        vertsBuffer.flip();

        vbo = glGenBuffers();
        vao = glGenVertexArrays();

        glBindVertexArray(vao);
        glBindBuffer(GL_ARRAY_BUFFER, vbo);
        glBufferData(GL_ARRAY_BUFFER, vertsBuffer , GL_STATIC_DRAW);

	}
	
	private Mat4 tmpMat = new Mat4();
	
	void preRender(float screenWidth, float screenHeight) {
		
		tmpMat.identity(); // screw with matrix to make it Cartesian 40x30 for 16x16char
		tmpMat.mat.put( 0, (2f/screenWidth)/(640/screenWidth));
		tmpMat.mat.put( 5,(-2f/screenHeight)/(480/screenHeight));
		tmpMat.mat.put(12,-1f);
		tmpMat.mat.put(13,1f);
		
		glUseProgram(program);
		
		glUniformMatrix4(u_matrix, false , tmpMat.mat);
		
		glBindVertexArray(vao);
		glBindBuffer(GL_ARRAY_BUFFER, vbo);

		glEnableVertexAttribArray(attr_pos);
		glEnableVertexAttribArray(attr_uv);

		// attrib, number items, type, normalised, byte stride, byte offset
		glVertexAttribPointer( attr_pos, 3, GL_FLOAT, false, 20, 0 );
		glVertexAttribPointer( attr_uv, 2, GL_FLOAT, false, 20, 12 );

		glBindTexture(GL_TEXTURE_2D, texture);
	}
	
	void render(float x, float y, String message) {
		
		int l = message.length();
		int c;
		float u,v;
		for (int n=0;n<l;n++) {
			c = message.charAt(n)-32;
			u = ((float)(c % 16)) * 0.0625f;
			v = (c/16) *0.1640625f;
			glUniform2f(u_offset, x+(n*charWidth), y);
			glUniform2f(u_uvoffset, u, v ); 
			glDrawArrays(GL_TRIANGLES, 0, 6);
		}
	}
	
	void postRender() {
		glDisableVertexAttribArray(attr_pos);
		glDisableVertexAttribArray(attr_uv);		
	}
	
}
